
# THIS FILE IS GENERATED FROM PYRAD_PROC SETUP.PY
short_version = '2.0.3'
version = '2.0.3'
full_version = '2.0.3'
git_revision = '7cb27317b16fa51ad6269d70739fd33d941fd413'
compile_date_time = '2025-05-20 07:35'
username = 'runner'
release = True

if not release:
    version = full_version
